@@ -2,7 +2,7 @@
 
   eval.c -
 
-  $Author: nagachika $
+  $Author: usa $
   created at: Thu Jun 10 14:22:17 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -458,6 +458,8 @@ exc_setup_cause(VALUE exc, VALUE cause)
     }
 #endif
     if (!NIL_P(cause) && cause != exc) {
+	if (OBJ_FROZEN(exc))
+	    exc = rb_obj_dup(exc);
 	rb_ivar_set(exc, id_cause, cause);
     }
     return exc;
@@ -469,6 +471,17 @@ sysstack_error_p(VALUE exc)
     return exc == sysstack_error || (!SPECIAL_CONST_P(exc) && RBASIC_CLASS(exc) == rb_eSysStackError);
 }
 
+static inline int
+special_exception_p(rb_thread_t *th, VALUE exc)
+{
+    enum ruby_special_exceptions i;
+    const VALUE *exceptions = th->vm->special_exceptions;
+    for (i = 0; i < ruby_special_error_count; ++i) {
+	if (exceptions[i] == exc) return TRUE;
+    }
+    return FALSE;
+}
+
 static void
 setup_exception(rb_thread_t *th, int tag, volatile VALUE mesg, VALUE cause)
 {
@@ -486,14 +499,17 @@ setup_exception(rb_thread_t *th, int tag, volatile VALUE mesg, VALUE cause)
 	mesg = rb_exc_new(rb_eRuntimeError, 0, 0);
 	nocause = 0;
     }
+    else if (special_exception_p(th, mesg)) {
+	mesg = ruby_vm_special_exception_copy(mesg);
+    }
     if (cause != Qundef) {
-	exc_setup_cause(mesg, cause);
+	mesg = exc_setup_cause(mesg, cause);
     }
     else if (nocause) {
-	exc_setup_cause(mesg, Qnil);
+	mesg = exc_setup_cause(mesg, Qnil);
     }
     else if (!rb_ivar_defined(mesg, id_cause)) {
-	exc_setup_cause(mesg, get_thread_errinfo(th));
+	mesg = exc_setup_cause(mesg, get_thread_errinfo(th));
     }
 
     file = rb_source_loc(&line);
@@ -502,9 +518,6 @@ setup_exception(rb_thread_t *th, int tag, volatile VALUE mesg, VALUE cause)
 	if (sysstack_error_p(mesg)) {
 	    if (NIL_P(rb_attr_get(mesg, idBt))) {
 		at = rb_vm_backtrace_object();
-		if (mesg == sysstack_error) {
-		    mesg = ruby_vm_sysstack_error_copy();
-		}
 		rb_ivar_set(mesg, idBt, at);
 		rb_ivar_set(mesg, idBt_locations, at);
 	    }
