@@ -1,3 +1,4 @@
+# frozen_string_literal: true
 ##
 # The global rubygems pool, available via the rubygems.org API.
 # Returns instances of APISpecification.
@@ -34,6 +35,8 @@ class Gem::Resolver::APISet < Gem::Resolver::Set
 
     @data   = Hash.new { |h,k| h[k] = [] }
     @source = Gem::Source.new @uri
+
+    @to_fetch = []
   end
 
   ##
@@ -45,6 +48,10 @@ class Gem::Resolver::APISet < Gem::Resolver::Set
 
     return res unless @remote
 
+    if @to_fetch.include?(req.name)
+      prefetch_now
+    end
+
     versions(req.name).each do |ver|
       if req.dependency.match? req.name, ver[:number]
         res << Gem::Resolver::APISpecification.new(self, ver)
@@ -61,9 +68,13 @@ class Gem::Resolver::APISet < Gem::Resolver::Set
   def prefetch reqs
     return unless @remote
     names = reqs.map { |r| r.dependency.name }
-    needed = names - @data.keys
+    needed = names - @data.keys - @to_fetch
 
-    return if needed.empty?
+    @to_fetch += needed
+  end
+
+  def prefetch_now # :nodoc:
+    needed, @to_fetch = @to_fetch, []
 
     uri = @dep_uri + "?gems=#{needed.sort.join ','}"
     str = Gem::RemoteFetcher.fetcher.fetch_path uri
