@@ -2,15 +2,15 @@
 
   math.c -
 
-  $Author: tadf $
+  $Author: usa $
   created at: Tue Jan 25 14:12:56 JST 1994
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
 
 **********************************************************************/
 
-#include "ruby/ruby.h"
 #include "internal.h"
+#include <float.h>
 #include <math.h>
 #include <errno.h>
 
@@ -19,26 +19,28 @@
     extern int signbit(double);
 #endif
 
-#define numberof(array) (int)(sizeof(array) / sizeof((array)[0]))
+#define RB_BIGNUM_TYPE_P(x) RB_TYPE_P((x), T_BIGNUM)
 
 VALUE rb_mMath;
 VALUE rb_eMathDomainError;
 
-#define Need_Float(x) do {if (!RB_TYPE_P(x, T_FLOAT)) {(x) = rb_to_float(x);}} while(0)
-#define Need_Float2(x,y) do {\
-    Need_Float(x);\
-    Need_Float(y);\
-} while (0)
+#define Get_Double(x) rb_num_to_dbl(x)
 
 #define domain_error(msg) \
     rb_raise(rb_eMathDomainError, "Numerical argument is out of domain - " #msg)
 
 /*
  *  call-seq:
- *     Math.atan2(y, x)  -> float
+ *     Math.atan2(y, x)  -> Float
  *
- *  Computes the arc tangent given <i>y</i> and <i>x</i>. Returns
- *  -PI..PI.
+ *  Computes the arc tangent given +y+ and +x+.
+ *  Returns a Float in the range -PI..PI. Return value is a angle
+ *  in radians between the positive x-axis of cartesian plane
+ *  and the point given by the coordinates (+x+, +y+) on it.
+ *
+ *  Domain: (-INFINITY, INFINITY)
+ *
+ *  Codomain: [-PI, PI]
  *
  *    Math.atan2(-0.0, -1.0) #=> -3.141592653589793
  *    Math.atan2(-1.0, -1.0) #=> -2.356194490192345
@@ -50,19 +52,19 @@ VALUE rb_eMathDomainError;
  *    Math.atan2(1.0, 0.0)   #=> 1.5707963267948966
  *    Math.atan2(1.0, -1.0)  #=> 2.356194490192345
  *    Math.atan2(0.0, -1.0)  #=> 3.141592653589793
+ *    Math.atan2(INFINITY, INFINITY)   #=> 0.7853981633974483
+ *    Math.atan2(INFINITY, -INFINITY)  #=> 2.356194490192345
+ *    Math.atan2(-INFINITY, INFINITY)  #=> -0.7853981633974483
+ *    Math.atan2(-INFINITY, -INFINITY) #=> -2.356194490192345
  *
  */
 
 static VALUE
 math_atan2(VALUE obj, VALUE y, VALUE x)
 {
-#ifndef M_PI
-# define M_PI 3.14159265358979323846
-#endif
     double dx, dy;
-    Need_Float2(y, x);
-    dx = RFLOAT_VALUE(x);
-    dy = RFLOAT_VALUE(y);
+    dx = Get_Double(x);
+    dy = Get_Double(y);
     if (dx == 0.0 && dy == 0.0) {
 	if (!signbit(dx))
 	    return DBL2NUM(dy);
@@ -70,110 +72,152 @@ math_atan2(VALUE obj, VALUE y, VALUE x)
 	    return DBL2NUM(M_PI);
 	return DBL2NUM(-M_PI);
     }
-    if (isinf(dx) && isinf(dy)) domain_error("atan2");
+#ifndef ATAN2_INF_C99
+    if (isinf(dx) && isinf(dy)) {
+	/* optimization for FLONUM */
+	if (dx < 0.0) {
+	    const double dz = (3.0 * M_PI / 4.0);
+	    return (dy < 0.0) ? DBL2NUM(-dz) : DBL2NUM(dz);
+	}
+	else {
+	    const double dz = (M_PI / 4.0);
+	    return (dy < 0.0) ? DBL2NUM(-dz) : DBL2NUM(dz);
+	}
+    }
+#endif
     return DBL2NUM(atan2(dy, dx));
 }
 
 
 /*
  *  call-seq:
- *     Math.cos(x)    -> float
+ *     Math.cos(x)    -> Float
+ *
+ *  Computes the cosine of +x+ (expressed in radians).
+ *  Returns a Float in the range -1.0..1.0.
+ *
+ *  Domain: (-INFINITY, INFINITY)
+ *
+ *  Codomain: [-1, 1]
+ *
+ *    Math.cos(Math::PI) #=> -1.0
  *
- *  Computes the cosine of <i>x</i> (expressed in radians). Returns
- *  -1..1.
  */
 
 static VALUE
 math_cos(VALUE obj, VALUE x)
 {
-    Need_Float(x);
-    return DBL2NUM(cos(RFLOAT_VALUE(x)));
+    return DBL2NUM(cos(Get_Double(x)));
 }
 
 /*
  *  call-seq:
- *     Math.sin(x)    -> float
+ *     Math.sin(x)    -> Float
+ *
+ *  Computes the sine of +x+ (expressed in radians).
+ *  Returns a Float in the range -1.0..1.0.
+ *
+ *  Domain: (-INFINITY, INFINITY)
+ *
+ *  Codomain: [-1, 1]
+ *
+ *    Math.sin(Math::PI/2) #=> 1.0
  *
- *  Computes the sine of <i>x</i> (expressed in radians). Returns
- *  -1..1.
  */
 
 static VALUE
 math_sin(VALUE obj, VALUE x)
 {
-    Need_Float(x);
-
-    return DBL2NUM(sin(RFLOAT_VALUE(x)));
+    return DBL2NUM(sin(Get_Double(x)));
 }
 
 
 /*
  *  call-seq:
- *     Math.tan(x)    -> float
+ *     Math.tan(x)    -> Float
+ *
+ *  Computes the tangent of +x+ (expressed in radians).
+ *
+ *  Domain: (-INFINITY, INFINITY)
+ *
+ *  Codomain: (-INFINITY, INFINITY)
+ *
+ *    Math.tan(0) #=> 0.0
  *
- *  Returns the tangent of <i>x</i> (expressed in radians).
  */
 
 static VALUE
 math_tan(VALUE obj, VALUE x)
 {
-    Need_Float(x);
-
-    return DBL2NUM(tan(RFLOAT_VALUE(x)));
+    return DBL2NUM(tan(Get_Double(x)));
 }
 
 /*
  *  call-seq:
- *     Math.acos(x)    -> float
+ *     Math.acos(x)    -> Float
+ *
+ *  Computes the arc cosine of +x+. Returns 0..PI.
+ *
+ *  Domain: [-1, 1]
+ *
+ *  Codomain: [0, PI]
+ *
+ *    Math.acos(0) == Math::PI/2  #=> true
  *
- *  Computes the arc cosine of <i>x</i>. Returns 0..PI.
  */
 
 static VALUE
 math_acos(VALUE obj, VALUE x)
 {
-    double d0, d;
+    double d;
 
-    Need_Float(x);
-    d0 = RFLOAT_VALUE(x);
+    d = Get_Double(x);
     /* check for domain error */
-    if (d0 < -1.0 || 1.0 < d0) domain_error("acos");
-    d = acos(d0);
-    return DBL2NUM(d);
+    if (d < -1.0 || 1.0 < d) domain_error("acos");
+    return DBL2NUM(acos(d));
 }
 
 /*
  *  call-seq:
- *     Math.asin(x)    -> float
+ *     Math.asin(x)    -> Float
  *
- *  Computes the arc sine of <i>x</i>. Returns -{PI/2} .. {PI/2}.
+ *  Computes the arc sine of +x+. Returns -PI/2..PI/2.
+ *
+ *  Domain: [-1, -1]
+ *
+ *  Codomain: [-PI/2, PI/2]
+ *
+ *    Math.asin(1) == Math::PI/2  #=> true
  */
 
 static VALUE
 math_asin(VALUE obj, VALUE x)
 {
-    double d0, d;
+    double d;
 
-    Need_Float(x);
-    d0 = RFLOAT_VALUE(x);
+    d = Get_Double(x);
     /* check for domain error */
-    if (d0 < -1.0 || 1.0 < d0) domain_error("asin");
-    d = asin(d0);
-    return DBL2NUM(d);
+    if (d < -1.0 || 1.0 < d) domain_error("asin");
+    return DBL2NUM(asin(d));
 }
 
 /*
  *  call-seq:
- *     Math.atan(x)    -> float
+ *     Math.atan(x)    -> Float
  *
- *  Computes the arc tangent of <i>x</i>. Returns -{PI/2} .. {PI/2}.
+ *  Computes the arc tangent of +x+. Returns -PI/2..PI/2.
+ *
+ *  Domain: (-INFINITY, INFINITY)
+ *
+ *  Codomain: (-PI/2, PI/2)
+ *
+ *    Math.atan(0) #=> 0.0
  */
 
 static VALUE
 math_atan(VALUE obj, VALUE x)
 {
-    Need_Float(x);
-    return DBL2NUM(atan(RFLOAT_VALUE(x)));
+    return DBL2NUM(atan(Get_Double(x)));
 }
 
 #ifndef HAVE_COSH
@@ -186,17 +230,22 @@ cosh(double x)
 
 /*
  *  call-seq:
- *     Math.cosh(x)    -> float
+ *     Math.cosh(x)    -> Float
+ *
+ *  Computes the hyperbolic cosine of +x+ (expressed in radians).
+ *
+ *  Domain: (-INFINITY, INFINITY)
+ *
+ *  Codomain: [1, INFINITY)
+ *
+ *    Math.cosh(0) #=> 1.0
  *
- *  Computes the hyperbolic cosine of <i>x</i> (expressed in radians).
  */
 
 static VALUE
 math_cosh(VALUE obj, VALUE x)
 {
-    Need_Float(x);
-
-    return DBL2NUM(cosh(RFLOAT_VALUE(x)));
+    return DBL2NUM(cosh(Get_Double(x)));
 }
 
 #ifndef HAVE_SINH
@@ -209,17 +258,22 @@ sinh(double x)
 
 /*
  *  call-seq:
- *     Math.sinh(x)    -> float
+ *     Math.sinh(x)    -> Float
+ *
+ *  Computes the hyperbolic sine of +x+ (expressed in radians).
+ *
+ *  Domain: (-INFINITY, INFINITY)
+ *
+ *  Codomain: (-INFINITY, INFINITY)
+ *
+ *    Math.sinh(0) #=> 0.0
  *
- *  Computes the hyperbolic sine of <i>x</i> (expressed in
- *  radians).
  */
 
 static VALUE
 math_sinh(VALUE obj, VALUE x)
 {
-    Need_Float(x);
-    return DBL2NUM(sinh(RFLOAT_VALUE(x)));
+    return DBL2NUM(sinh(Get_Double(x)));
 }
 
 #ifndef HAVE_TANH
@@ -232,82 +286,107 @@ tanh(double x)
 
 /*
  *  call-seq:
- *     Math.tanh()    -> float
+ *     Math.tanh(x)    -> Float
+ *
+ *  Computes the hyperbolic tangent of +x+ (expressed in radians).
+ *
+ *  Domain: (-INFINITY, INFINITY)
+ *
+ *  Codomain: (-1, 1)
+ *
+ *    Math.tanh(0) #=> 0.0
  *
- *  Computes the hyperbolic tangent of <i>x</i> (expressed in
- *  radians).
  */
 
 static VALUE
 math_tanh(VALUE obj, VALUE x)
 {
-    Need_Float(x);
-    return DBL2NUM(tanh(RFLOAT_VALUE(x)));
+    return DBL2NUM(tanh(Get_Double(x)));
 }
 
 /*
  *  call-seq:
- *     Math.acosh(x)    -> float
+ *     Math.acosh(x)    -> Float
+ *
+ *  Computes the inverse hyperbolic cosine of +x+.
+ *
+ *  Domain: [1, INFINITY)
+ *
+ *  Codomain: [0, INFINITY)
+ *
+ *    Math.acosh(1) #=> 0.0
  *
- *  Computes the inverse hyperbolic cosine of <i>x</i>.
  */
 
 static VALUE
 math_acosh(VALUE obj, VALUE x)
 {
-    double d0, d;
+    double d;
 
-    Need_Float(x);
-    d0 = RFLOAT_VALUE(x);
+    d = Get_Double(x);
     /* check for domain error */
-    if (d0 < 1.0) domain_error("acosh");
-    d = acosh(d0);
-    return DBL2NUM(d);
+    if (d < 1.0) domain_error("acosh");
+    return DBL2NUM(acosh(d));
 }
 
 /*
  *  call-seq:
- *     Math.asinh(x)    -> float
+ *     Math.asinh(x)    -> Float
+ *
+ *  Computes the inverse hyperbolic sine of +x+.
+ *
+ *  Domain: (-INFINITY, INFINITY)
+ *
+ *  Codomain: (-INFINITY, INFINITY)
+ *
+ *    Math.asinh(1) #=> 0.881373587019543
  *
- *  Computes the inverse hyperbolic sine of <i>x</i>.
  */
 
 static VALUE
 math_asinh(VALUE obj, VALUE x)
 {
-    Need_Float(x);
-    return DBL2NUM(asinh(RFLOAT_VALUE(x)));
+    return DBL2NUM(asinh(Get_Double(x)));
 }
 
 /*
  *  call-seq:
- *     Math.atanh(x)    -> float
+ *     Math.atanh(x)    -> Float
+ *
+ *  Computes the inverse hyperbolic tangent of +x+.
+ *
+ *  Domain: (-1, 1)
+ *
+ *  Codomain: (-INFINITY, INFINITY)
+ *
+ *   Math.atanh(1) #=> Infinity
  *
- *  Computes the inverse hyperbolic tangent of <i>x</i>.
  */
 
 static VALUE
 math_atanh(VALUE obj, VALUE x)
 {
-    double d0, d;
+    double d;
 
-    Need_Float(x);
-    d0 = RFLOAT_VALUE(x);
+    d = Get_Double(x);
     /* check for domain error */
-    if (d0 <  -1.0 || +1.0 <  d0) domain_error("atanh");
+    if (d <  -1.0 || +1.0 <  d) domain_error("atanh");
     /* check for pole error */
-    if (d0 == -1.0) return DBL2NUM(-INFINITY);
-    if (d0 == +1.0) return DBL2NUM(+INFINITY);
-    d = atanh(d0);
-    return DBL2NUM(d);
+    if (d == -1.0) return DBL2NUM(-INFINITY);
+    if (d == +1.0) return DBL2NUM(+INFINITY);
+    return DBL2NUM(atanh(d));
 }
 
 /*
  *  call-seq:
- *     Math.exp(x)    -> float
+ *     Math.exp(x)    -> Float
  *
  *  Returns e**x.
  *
+ *  Domain: (-INFINITY, INFINITY)
+ *
+ *  Codomain: (0, INFINITY)
+ *
  *    Math.exp(0)       #=> 1.0
  *    Math.exp(1)       #=> 2.718281828459045
  *    Math.exp(1.5)     #=> 4.4816890703380645
@@ -317,8 +396,7 @@ math_atanh(VALUE obj, VALUE x)
 static VALUE
 math_exp(VALUE obj, VALUE x)
 {
-    Need_Float(x);
-    return DBL2NUM(exp(RFLOAT_VALUE(x)));
+    return DBL2NUM(exp(Get_Double(x)));
 }
 
 #if defined __CYGWIN__
@@ -330,43 +408,67 @@ math_exp(VALUE obj, VALUE x)
 # define log10(x) ((x) < 0.0 ? nan("") : log10(x))
 #endif
 
+static double math_log1(VALUE x);
+
 /*
  *  call-seq:
- *     Math.log(numeric)    -> float
- *     Math.log(num,base)   -> float
+ *     Math.log(x)          -> Float
+ *     Math.log(x, base)    -> Float
  *
- *  Returns the natural logarithm of <i>numeric</i>.
+ *  Returns the logarithm of +x+.
  *  If additional second argument is given, it will be the base
- *  of logarithm.
+ *  of logarithm. Otherwise it is +e+ (for the natural logarithm).
  *
+ *  Domain: (0, INFINITY)
+ *
+ *  Codomain: (-INFINITY, INFINITY)
+ *
+ *    Math.log(0)          #=> -Infinity
  *    Math.log(1)          #=> 0.0
  *    Math.log(Math::E)    #=> 1.0
  *    Math.log(Math::E**3) #=> 3.0
- *    Math.log(12,3)       #=> 2.2618595071429146
+ *    Math.log(12, 3)      #=> 2.2618595071429146
  *
  */
 
 static VALUE
-math_log(int argc, VALUE *argv)
+math_log(int argc, const VALUE *argv, VALUE obj)
 {
     VALUE x, base;
-    double d0, d;
+    double d;
 
     rb_scan_args(argc, argv, "11", &x, &base);
-    Need_Float(x);
-    d0 = RFLOAT_VALUE(x);
-    /* check for domain error */
-    if (d0 < 0.0) domain_error("log");
-    /* check for pole error */
-    if (d0 == 0.0) return DBL2NUM(-INFINITY);
-    d = log(d0);
+    d = math_log1(x);
     if (argc == 2) {
-	Need_Float(base);
-	d /= log(RFLOAT_VALUE(base));
+	d /= math_log1(base);
     }
     return DBL2NUM(d);
 }
 
+static double
+math_log1(VALUE x)
+{
+    double d;
+    size_t numbits;
+
+    if (RB_BIGNUM_TYPE_P(x) && BIGNUM_POSITIVE_P(x) &&
+            DBL_MAX_EXP <= (numbits = rb_absint_numwords(x, 1, NULL))) {
+        numbits -= DBL_MANT_DIG;
+        x = rb_big_rshift(x, SIZET2NUM(numbits));
+    }
+    else {
+	numbits = 0;
+    }
+
+    d = Get_Double(x);
+    /* check for domain error */
+    if (d < 0.0) domain_error("log");
+    /* check for pole error */
+    if (d == 0.0) return -INFINITY;
+
+    return log(d) + numbits * log(2); /* log(d * 2 ** numbits) */
+}
+
 #ifndef log2
 #ifndef HAVE_LOG2
 double
@@ -381,9 +483,13 @@ extern double log2(double);
 
 /*
  *  call-seq:
- *     Math.log2(numeric)    -> float
+ *     Math.log2(x)    -> Float
  *
- *  Returns the base 2 logarithm of <i>numeric</i>.
+ *  Returns the base 2 logarithm of +x+.
+ *
+ *  Domain: (0, INFINITY)
+ *
+ *  Codomain: (-INFINITY, INFINITY)
  *
  *    Math.log2(1)      #=> 0.0
  *    Math.log2(2)      #=> 1.0
@@ -395,23 +501,36 @@ extern double log2(double);
 static VALUE
 math_log2(VALUE obj, VALUE x)
 {
-    double d0, d;
+    double d;
+    size_t numbits;
 
-    Need_Float(x);
-    d0 = RFLOAT_VALUE(x);
+    if (RB_BIGNUM_TYPE_P(x) && BIGNUM_POSITIVE_P(x) &&
+            DBL_MAX_EXP <= (numbits = rb_absint_numwords(x, 1, NULL))) {
+        numbits -= DBL_MANT_DIG;
+        x = rb_big_rshift(x, SIZET2NUM(numbits));
+    }
+    else {
+	numbits = 0;
+    }
+
+    d = Get_Double(x);
     /* check for domain error */
-    if (d0 < 0.0) domain_error("log2");
+    if (d < 0.0) domain_error("log2");
     /* check for pole error */
-    if (d0 == 0.0) return DBL2NUM(-INFINITY);
-    d = log2(d0);
-    return DBL2NUM(d);
+    if (d == 0.0) return DBL2NUM(-INFINITY);
+
+    return DBL2NUM(log2(d) + numbits); /* log2(d * 2 ** numbits) */
 }
 
 /*
  *  call-seq:
- *     Math.log10(numeric)    -> float
+ *     Math.log10(x)    -> Float
  *
- *  Returns the base 10 logarithm of <i>numeric</i>.
+ *  Returns the base 10 logarithm of +x+.
+ *
+ *  Domain: (0, INFINITY)
+ *
+ *  Codomain: (-INFINITY, INFINITY)
  *
  *    Math.log10(1)       #=> 0.0
  *    Math.log10(10)      #=> 1.0
@@ -422,102 +541,112 @@ math_log2(VALUE obj, VALUE x)
 static VALUE
 math_log10(VALUE obj, VALUE x)
 {
-    double d0, d;
+    double d;
+    size_t numbits;
 
-    Need_Float(x);
-    d0 = RFLOAT_VALUE(x);
+    if (RB_BIGNUM_TYPE_P(x) && BIGNUM_POSITIVE_P(x) &&
+            DBL_MAX_EXP <= (numbits = rb_absint_numwords(x, 1, NULL))) {
+        numbits -= DBL_MANT_DIG;
+        x = rb_big_rshift(x, SIZET2NUM(numbits));
+    }
+    else {
+	numbits = 0;
+    }
+
+    d = Get_Double(x);
     /* check for domain error */
-    if (d0 < 0.0) domain_error("log10");
+    if (d < 0.0) domain_error("log10");
     /* check for pole error */
-    if (d0 == 0.0) return DBL2NUM(-INFINITY);
-    d = log10(d0);
-    return DBL2NUM(d);
+    if (d == 0.0) return DBL2NUM(-INFINITY);
+
+    return DBL2NUM(log10(d) + numbits * log10(2)); /* log10(d * 2 ** numbits) */
 }
 
 /*
  *  call-seq:
- *     Math.sqrt(numeric)    -> float
+ *     Math.sqrt(x)    -> Float
  *
- *  Returns the non-negative square root of <i>numeric</i>.
+ *  Returns the non-negative square root of +x+.
+ *
+ *  Domain: [0, INFINITY)
+ *
+ *  Codomain:[0, INFINITY)
  *
  *    0.upto(10) {|x|
  *      p [x, Math.sqrt(x), Math.sqrt(x)**2]
  *    }
- *    #=>
- *    [0, 0.0, 0.0]
- *    [1, 1.0, 1.0]
- *    [2, 1.4142135623731, 2.0]
- *    [3, 1.73205080756888, 3.0]
- *    [4, 2.0, 4.0]
- *    [5, 2.23606797749979, 5.0]
- *    [6, 2.44948974278318, 6.0]
- *    [7, 2.64575131106459, 7.0]
- *    [8, 2.82842712474619, 8.0]
- *    [9, 3.0, 9.0]
- *    [10, 3.16227766016838, 10.0]
- *
+ *    #=> [0, 0.0, 0.0]
+ *    #   [1, 1.0, 1.0]
+ *    #   [2, 1.4142135623731, 2.0]
+ *    #   [3, 1.73205080756888, 3.0]
+ *    #   [4, 2.0, 4.0]
+ *    #   [5, 2.23606797749979, 5.0]
+ *    #   [6, 2.44948974278318, 6.0]
+ *    #   [7, 2.64575131106459, 7.0]
+ *    #   [8, 2.82842712474619, 8.0]
+ *    #   [9, 3.0, 9.0]
+ *    #   [10, 3.16227766016838, 10.0]
  */
 
 static VALUE
 math_sqrt(VALUE obj, VALUE x)
 {
-    double d0, d;
+    double d;
 
-    Need_Float(x);
-    d0 = RFLOAT_VALUE(x);
+    d = Get_Double(x);
     /* check for domain error */
-    if (d0 < 0.0) domain_error("sqrt");
-    if (d0 == 0.0) return DBL2NUM(0.0);
-    d = sqrt(d0);
-    return DBL2NUM(d);
+    if (d < 0.0) domain_error("sqrt");
+    if (d == 0.0) return DBL2NUM(0.0);
+    return DBL2NUM(sqrt(d));
 }
 
 /*
  *  call-seq:
- *     Math.cbrt(numeric)    -> float
+ *     Math.cbrt(x)    -> Float
  *
- *  Returns the cube root of <i>numeric</i>.
+ *  Returns the cube root of +x+.
+ *
+ *  Domain: (-INFINITY, INFINITY)
+ *
+ *  Codomain: (-INFINITY, INFINITY)
  *
  *    -9.upto(9) {|x|
  *      p [x, Math.cbrt(x), Math.cbrt(x)**3]
  *    }
- *    #=>
- *    [-9, -2.0800838230519, -9.0]
- *    [-8, -2.0, -8.0]
- *    [-7, -1.91293118277239, -7.0]
- *    [-6, -1.81712059283214, -6.0]
- *    [-5, -1.7099759466767, -5.0]
- *    [-4, -1.5874010519682, -4.0]
- *    [-3, -1.44224957030741, -3.0]
- *    [-2, -1.25992104989487, -2.0]
- *    [-1, -1.0, -1.0]
- *    [0, 0.0, 0.0]
- *    [1, 1.0, 1.0]
- *    [2, 1.25992104989487, 2.0]
- *    [3, 1.44224957030741, 3.0]
- *    [4, 1.5874010519682, 4.0]
- *    [5, 1.7099759466767, 5.0]
- *    [6, 1.81712059283214, 6.0]
- *    [7, 1.91293118277239, 7.0]
- *    [8, 2.0, 8.0]
- *    [9, 2.0800838230519, 9.0]
+ *    #=> [-9, -2.0800838230519, -9.0]
+ *    #   [-8, -2.0, -8.0]
+ *    #   [-7, -1.91293118277239, -7.0]
+ *    #   [-6, -1.81712059283214, -6.0]
+ *    #   [-5, -1.7099759466767, -5.0]
+ *    #   [-4, -1.5874010519682, -4.0]
+ *    #   [-3, -1.44224957030741, -3.0]
+ *    #   [-2, -1.25992104989487, -2.0]
+ *    #   [-1, -1.0, -1.0]
+ *    #   [0, 0.0, 0.0]
+ *    #   [1, 1.0, 1.0]
+ *    #   [2, 1.25992104989487, 2.0]
+ *    #   [3, 1.44224957030741, 3.0]
+ *    #   [4, 1.5874010519682, 4.0]
+ *    #   [5, 1.7099759466767, 5.0]
+ *    #   [6, 1.81712059283214, 6.0]
+ *    #   [7, 1.91293118277239, 7.0]
+ *    #   [8, 2.0, 8.0]
+ *    #   [9, 2.0800838230519, 9.0]
  *
  */
 
 static VALUE
 math_cbrt(VALUE obj, VALUE x)
 {
-    Need_Float(x);
-    return DBL2NUM(cbrt(RFLOAT_VALUE(x)));
+    return DBL2NUM(cbrt(Get_Double(x)));
 }
 
 /*
  *  call-seq:
- *     Math.frexp(numeric)    -> [ fraction, exponent ]
+ *     Math.frexp(x)    -> [fraction, exponent]
  *
- *  Returns a two-element array containing the normalized fraction (a
- *  <code>Float</code>) and exponent (a <code>Fixnum</code>) of
- *  <i>numeric</i>.
+ *  Returns a two-element array containing the normalized fraction (a Float)
+ *  and exponent (a Fixnum) of +x+.
  *
  *     fraction, exponent = Math.frexp(1234)   #=> [0.6025390625, 11]
  *     fraction * 2**exponent                  #=> 1234.0
@@ -529,17 +658,15 @@ math_frexp(VALUE obj, VALUE x)
     double d;
     int exp;
 
-    Need_Float(x);
-
-    d = frexp(RFLOAT_VALUE(x), &exp);
+    d = frexp(Get_Double(x), &exp);
     return rb_assoc_new(DBL2NUM(d), INT2NUM(exp));
 }
 
 /*
  *  call-seq:
- *     Math.ldexp(flt, int) -> float
+ *     Math.ldexp(fraction, exponent) -> float
  *
- *  Returns the value of <i>flt</i>*(2**<i>int</i>).
+ *  Returns the value of +fraction+*(2**+exponent+).
  *
  *     fraction, exponent = Math.frexp(1234)
  *     Math.ldexp(fraction, exponent)   #=> 1234.0
@@ -548,16 +675,15 @@ math_frexp(VALUE obj, VALUE x)
 static VALUE
 math_ldexp(VALUE obj, VALUE x, VALUE n)
 {
-    Need_Float(x);
-    return DBL2NUM(ldexp(RFLOAT_VALUE(x), NUM2INT(n)));
+    return DBL2NUM(ldexp(Get_Double(x), NUM2INT(n)));
 }
 
 /*
  *  call-seq:
- *     Math.hypot(x, y)    -> float
+ *     Math.hypot(x, y)    -> Float
  *
- *  Returns sqrt(x**2 + y**2), the hypotenuse of a right-angled triangle
- *  with sides <i>x</i> and <i>y</i>.
+ *  Returns sqrt(x**2 + y**2), the hypotenuse of a right-angled triangle with
+ *  sides +x+ and +y+.
  *
  *     Math.hypot(3, 4)   #=> 5.0
  */
@@ -565,41 +691,87 @@ math_ldexp(VALUE obj, VALUE x, VALUE n)
 static VALUE
 math_hypot(VALUE obj, VALUE x, VALUE y)
 {
-    Need_Float2(x, y);
-    return DBL2NUM(hypot(RFLOAT_VALUE(x), RFLOAT_VALUE(y)));
+    return DBL2NUM(hypot(Get_Double(x), Get_Double(y)));
 }
 
 /*
  * call-seq:
- *    Math.erf(x)  -> float
+ *    Math.erf(x)  -> Float
+ *
+ *  Calculates the error function of +x+.
+ *
+ *  Domain: (-INFINITY, INFINITY)
+ *
+ *  Codomain: (-1, 1)
+ *
+ *    Math.erf(0) #=> 0.0
  *
- *  Calculates the error function of x.
  */
 
 static VALUE
 math_erf(VALUE obj, VALUE x)
 {
-    Need_Float(x);
-    return DBL2NUM(erf(RFLOAT_VALUE(x)));
+    return DBL2NUM(erf(Get_Double(x)));
 }
 
 /*
  * call-seq:
- *    Math.erfc(x)  -> float
+ *    Math.erfc(x)  -> Float
  *
  *  Calculates the complementary error function of x.
+ *
+ *  Domain: (-INFINITY, INFINITY)
+ *
+ *  Codomain: (0, 2)
+ *
+ *    Math.erfc(0) #=> 1.0
+ *
  */
 
 static VALUE
 math_erfc(VALUE obj, VALUE x)
 {
-    Need_Float(x);
-    return DBL2NUM(erfc(RFLOAT_VALUE(x)));
+    return DBL2NUM(erfc(Get_Double(x)));
 }
 
+#if defined __MINGW32__
+static inline double
+ruby_tgamma(const double d)
+{
+    const double g = tgamma(d);
+    if (isinf(g)) {
+	if (d == 0.0 && signbit(d)) return -INFINITY;
+    }
+    if (isnan(g)) {
+	if (!signbit(d)) return INFINITY;
+    }
+    return g;
+}
+#define tgamma(d) ruby_tgamma(d)
+#endif
+
+#if defined LGAMMA_R_PM0_FIX
+static inline double
+ruby_lgamma_r(const double d, int *sign)
+{
+    const double g = lgamma_r(d, sign);
+    if (isinf(g)) {
+	if (d == 0.0 && signbit(d)) {
+	    *sign = -1;
+	    return INFINITY;
+	} else if (d == 0.0 && !signbit(d)) {
+	    *sign = 1;
+	    return INFINITY;
+	}
+    }
+    return g;
+}
+#define lgamma_r(d, sign) ruby_lgamma_r(d, sign)
+#endif
+
 /*
  * call-seq:
- *    Math.gamma(x)  -> float
+ *    Math.gamma(x)  -> Float
  *
  *  Calculates the gamma function of x.
  *
@@ -668,51 +840,47 @@ math_gamma(VALUE obj, VALUE x)
          * impossible to represent exactly in IEEE 754 double which have
          * 53bit mantissa. */
     };
-    double d0, d;
-    double intpart, fracpart;
-    Need_Float(x);
-    d0 = RFLOAT_VALUE(x);
+    enum {NFACT_TABLE = numberof(fact_table)};
+    double d;
+    d = Get_Double(x);
     /* check for domain error */
-    if (isinf(d0) && signbit(d0)) domain_error("gamma");
-    fracpart = modf(d0, &intpart);
-    if (fracpart == 0.0) {
-	if (intpart < 0) domain_error("gamma");
-	if (0 < intpart &&
-	    intpart - 1 < (double)numberof(fact_table)) {
-	    return DBL2NUM(fact_table[(int)intpart - 1]);
+    if (isinf(d) && signbit(d)) domain_error("gamma");
+    if (d == floor(d)) {
+	if (d < 0.0) domain_error("gamma");
+	if (1.0 <= d && d <= (double)NFACT_TABLE) {
+	    return DBL2NUM(fact_table[(int)d - 1]);
 	}
     }
-    d = tgamma(d0);
-    return DBL2NUM(d);
+    return DBL2NUM(tgamma(d));
 }
 
 /*
  * call-seq:
  *    Math.lgamma(x)  -> [float, -1 or 1]
  *
- *  Calculates the logarithmic gamma of x and
- *  the sign of gamma of x.
+ *  Calculates the logarithmic gamma of +x+ and the sign of gamma of +x+.
  *
  *  Math.lgamma(x) is same as
  *   [Math.log(Math.gamma(x).abs), Math.gamma(x) < 0 ? -1 : 1]
  *  but avoid overflow by Math.gamma(x) for large x.
+ *
+ *    Math.lgamma(0) #=> [Infinity, 1]
+ *
  */
 
 static VALUE
 math_lgamma(VALUE obj, VALUE x)
 {
-    double d0, d;
+    double d;
     int sign=1;
     VALUE v;
-    Need_Float(x);
-    d0 = RFLOAT_VALUE(x);
+    d = Get_Double(x);
     /* check for domain error */
-    if (isinf(d0)) {
-	if (signbit(d0)) domain_error("lgamma");
+    if (isinf(d)) {
+	if (signbit(d)) domain_error("lgamma");
 	return rb_assoc_new(DBL2NUM(INFINITY), INT2FIX(1));
     }
-    d = lgamma_r(d0, &sign);
-    v = DBL2NUM(d);
+    v = DBL2NUM(lgamma_r(d, &sign));
     return rb_assoc_new(v, INT2FIX(sign));
 }
 
@@ -738,14 +906,16 @@ exp1(exp)
 exp2(hypot)
 
 VALUE
-rb_math_log(int argc, VALUE *argv)
+rb_math_log(int argc, const VALUE *argv)
 {
-    return math_log(argc, argv);
+    return math_log(argc, argv, rb_mMath);
 }
 
 exp1(sin)
 exp1(sinh)
+#if 0
 exp1(sqrt)
+#endif
 
 
 /*
@@ -767,26 +937,30 @@ exp1(sqrt)
 /*
  *  Document-class: Math
  *
- *  The <code>Math</code> module contains module functions for basic
+ *  The Math module contains module functions for basic
  *  trigonometric and transcendental functions. See class
- *  <code>Float</code> for a list of constants that
+ *  Float for a list of constants that
  *  define Ruby's floating point accuracy.
+ *
+ *  Domains and codomains are given only for real (not complex) numbers.
  */
 
 
 void
-Init_Math(void)
+InitVM_Math(void)
 {
     rb_mMath = rb_define_module("Math");
     rb_eMathDomainError = rb_define_class_under(rb_mMath, "DomainError", rb_eStandardError);
 
 #ifdef M_PI
+    /*  Definition of the mathematical constant PI as a Float number. */
     rb_define_const(rb_mMath, "PI", DBL2NUM(M_PI));
 #else
     rb_define_const(rb_mMath, "PI", DBL2NUM(atan(1.0)*4.0));
 #endif
 
 #ifdef M_E
+    /*  Definition of the mathematical constant E (e) as a Float number. */
     rb_define_const(rb_mMath, "E", DBL2NUM(M_E));
 #else
     rb_define_const(rb_mMath, "E", DBL2NUM(exp(1.0)));
@@ -827,3 +1001,9 @@ Init_Math(void)
     rb_define_module_function(rb_mMath, "gamma", math_gamma, 1);
     rb_define_module_function(rb_mMath, "lgamma", math_lgamma, 1);
 }
+
+void
+Init_Math(void)
+{
+    InitVM(Math);
+}
