@@ -1,3 +1,4 @@
+# frozen_string_literal: false
 require 'test/unit'
 
 ConstTest = 3
@@ -7,7 +8,7 @@ class Class
 end
 
 class TestBasicInstructions < Test::Unit::TestCase
-  
+
   def test_immediates
     assert_equal((1==1), true)
     assert_equal((1==2), false)
@@ -30,7 +31,7 @@ class TestBasicInstructions < Test::Unit::TestCase
     assert_equal false, (self == nil)
     assert_equal false, (self == 0)
   end
-  
+
   def test_string
     expected = "str" + "ing"
     assert_equal expected, 'string'
@@ -55,16 +56,16 @@ class TestBasicInstructions < Test::Unit::TestCase
     assert_equal :sym, :"#{s}"
     assert_equal :sym, :"#{"#{"#{s}"}"}"
   end
-  
+
   def test_xstr
     assert_equal 'hoge', `echo hoge`.chomp
     assert_equal '3', `echo #{1 + 2}`.chomp
     hoge = 'huga'
     assert_equal 'huga', `echo #{hoge}`.chomp
   end
-  
+
   def test_regexp
-    assert_equal /test/, /test/
+    assert_equal(/test/, /test/)
     assert_equal 'test', /test/.source
     assert_equal 'TEST', /TEST/.source
     assert_equal true, !!(/test/ =~ 'test')
@@ -76,16 +77,18 @@ class TestBasicInstructions < Test::Unit::TestCase
     assert_equal true, !!(re =~ 'test')
     assert_equal false, !!(re =~ 'does not match')
 
-    assert_equal /x#{1+1}x/, /x#{1+1}x/
+    assert_equal(/x#{1+1}x/, /x#{1+1}x/)
     s = "OK"
-    assert_equal /x#{s}x/, /x#{s}x/
+    assert_equal(/x#{s}x/, /x#{s}x/)
     assert_equal true, !!(/x#{s}x/ =~ "xOKx")
     assert_equal false, !!(/x#{s}x/ =~ "does not match")
 
     s = "OK"
     prev = nil
     3.times do
-      assert_equal prev.object_id, (prev ||= /#{s}/o).object_id  if prev
+      re = /#{s}/o
+      assert_same prev, re if prev
+      prev = re
     end
   end
 
@@ -108,7 +111,7 @@ class TestBasicInstructions < Test::Unit::TestCase
     assert_equal 'c', a[2]
     assert_nil a[3]
   end
-  
+
   def test_hash
     assert_equal({}, {})
     assert_equal({1=>2}, {1=>2})
@@ -127,7 +130,7 @@ class TestBasicInstructions < Test::Unit::TestCase
     assert_equal((1...3), (1...2+1))
     assert_equal(('a'..'z'), ('a'..'z'))
   end
-  
+
   def test_not
     assert_equal true, !nil
     assert_equal true, !false
@@ -500,6 +503,7 @@ class TestBasicInstructions < Test::Unit::TestCase
 
   class OP
     attr_reader :x
+    attr_accessor :foo
     def x=(x)
       @x = x
       :Bug1996
@@ -600,6 +604,19 @@ class TestBasicInstructions < Test::Unit::TestCase
     assert_equal 4, x[0]
   end
 
+  def test_send_opassign
+    return if defined?(RUBY_ENGINE) and RUBY_ENGINE != "ruby"
+
+    bug7773 = '[ruby-core:51821]'
+    x = OP.new
+    assert_equal 42, x.foo = 42, bug7773
+    assert_equal 42, x.foo, bug7773
+    assert_equal -6, x.send(:foo=, -6), bug7773
+    assert_equal -6, x.foo, bug7773
+    assert_equal :Bug1996, x.send(:x=, :case_when_setter_returns_other_value), bug7773
+    assert_equal :case_when_setter_returns_other_value, x.x, bug7773
+  end
+
   def test_backref
     /re/ =~ 'not match'
     assert_nil $~
@@ -632,7 +649,7 @@ class TestBasicInstructions < Test::Unit::TestCase
     assert_equal 'i', $~[9]
     assert_equal 'x', $`
     assert_equal 'abcdefghi', $&
-    assert_equal 'y', $'
+    assert_equal "y", $'
     assert_equal 'i', $+
     assert_equal 'a', $1
     assert_equal 'b', $2
@@ -662,19 +679,23 @@ class TestBasicInstructions < Test::Unit::TestCase
   end
 
   def test_array_splat
+    feature1125 = '[ruby-core:21901]'
+
     a = []
     assert_equal [], [*a]
     assert_equal [1], [1, *a]
+    assert_not_same(a, [*a], feature1125)
     a = [2]
     assert_equal [2], [*a]
     assert_equal [1, 2], [1, *a]
+    assert_not_same(a, [*a], feature1125)
     a = [2, 3]
     assert_equal [2, 3], [*a]
     assert_equal [1, 2, 3], [1, *a]
+    assert_not_same(a, [*a], feature1125)
 
     a = nil
     assert_equal [], [*a]
     assert_equal [1], [1, *a]
   end
-
 end
